#!/usr/bin/env node

const fs = require('fs');
const child_process = require('child_process');
const { promisify } = require('util');
const { EOL } = require('os');

const exec = promisify(child_process.exec);
const timeToWrite = 5000;
const taskContract = /[A-Z]+-[0-9]+/;
const taskTypeContact = /(feature|bugfix|hotfix|doc|refactor)/;
const branchContract = /^((feature|bugfix|hotfix)\/[A-Z]+-[0-9]+\/.*|(doc|refactor)\/.*)/;
const notTestingBranchesContracts = /^(master|dev|test|stage|release-\d*\.?\d*|hotfix-\d*\.?\d*)/;
const commitEditMsgFile = process.env.HUSKY_GIT_PARAMS || process.argv[2];

const FORMAT = {
    RESET: '\x1b[0m',
    BOLD: '\x1b[1m',
    UNDERLINE: '\x1b[4m',
    NEW_LINE: '\n',
    DOUBLE_NEW_LINE: '\n\n',
};

const COLORS = {
    RED: '\x1b[31m',
    ORANGE: '\x1b[46m',
    GREEN: '\x1b[32m',
    BGGREEN: '\x1b[42m',
    BGMAGENTA: '\x1b[43m',
    BLUE: '\u001b[34m',
};

const logger = {
    default: (text) => console.log(FORMAT.RESET, text, FORMAT.RESET),
    success: (text) => console.log(COLORS.GREEN, text, FORMAT.RESET),
    orange: (text) => console.log(COLORS.ORANGE, text, FORMAT.RESET),
    bgGreen: (text) => console.log(COLORS.BGGREEN, text, FORMAT.RESET),
    bgMagenta: (text) => console.log(COLORS.BGMAGENTA, text, FORMAT.RESET),
    error: (text) => console.error(COLORS.RED, text, FORMAT.RESET),
    blue: (text) => console.error(COLORS.BLUE, text, FORMAT.RESET),
};

async function getCurrentBranch() {
    const branchOutput = await exec('git symbolic-ref --short HEAD');
    const { stdout, stderr } = branchOutput;

    if (stderr.length) {
        throw new Error(stderr);
    }

    return stdout.trim();
}

function getCommitMessageArray(file) {
    const message = fs.readFileSync(file, 'utf8');

    return message.split(EOL);
}

async function checkBranch(branchName) {
    if (notTestingBranchesContracts.test(branchName)) {
        return true;
    }

    if (!branchContract.test(branchName)) {
        logger.error('');
        console.error(`⚠️  \x1b[1;31mНазвание ветки не соответствует шаблону${FORMAT.RESET}\n`);
        console.log('Переименуйте ветку: git branch -m "feature/CBDL-123/someFeature \n');
        console.log('\x1b[32mПримеры имен веток:');
        console.log(`\x1b[32m------------------------------------------|${FORMAT.RESET}`);
        console.log('doc/add_docblocks_to_all_classes \x1b[32m         |', FORMAT.RESET);
        console.log('feature/CBDL-635/create_deposite \x1b[32m         |', FORMAT.RESET);
        console.log('feature/CBDL-635/schedule_payment \x1b[32m        |', FORMAT.RESET);
        console.log('hotfix/CBDL-678/remove_spaces_from_login \x1b[32m |', FORMAT.RESET);
        console.log(`\x1b[32m------------------------------------------|${FORMAT.RESET}\n`);
        console.log(`Подробнее: ${COLORS.BLUE}${FORMAT.UNDERLINE}https://kb.akbars.ru/pages/viewpage.action?pageId=73574926${FORMAT.RESET}`);
        logger.default('');

        return Promise.reject(new Error(''));
    }

    return false;
}

async function getTaskFromBranch(branchName) {
    try {
        await checkBranch(branchName);

        const [_, task] = branchName.split('/');

        if (task) {
            return task.replace(/\s+/g, '');
        }
    } catch (error) {
        throw new Error(error.message);
    }
}

async function getTaskTypeFromBranch(branchName) {
    try {
        const notTestingBranch = await checkBranch(branchName);

        if (notTestingBranch) {
            return null;
        }

        const [taskType] = branchName.split('/');

        if (taskType) {
            return taskType.replace(/\s+/g, '');
        }
    } catch (error) {
        throw new Error(error.message);
    }
}

function existTaskInFile(file) {
    const lines = getCommitMessageArray(file);
    const withoutComments = lines.filter(l => taskContract.test(l)).join('');

    return taskContract.test(withoutComments);
}

function existTaskTypeInFile(file) {
    const lines = getCommitMessageArray(file);
    const withoutComments = lines.filter(l => taskTypeContact.test(l)).join('');

    return taskTypeContact.test(withoutComments);
}

function writeTaskToTitleInFile(task, file) {
    const lines = getCommitMessageArray(file);
    lines[0] = `#${task} ${lines[0]}`;
    const newLines = lines.join(EOL);

    fs.writeFileSync(file, newLines, 'utf8');
}

function writeTaskTypeToTitleInFile(taskType, file) {
    const lines = getCommitMessageArray(file);
    lines[0] = `[${taskType}] ${lines[0]}`;
    const newLines = lines.join(EOL);

    fs.writeFileSync(file, newLines, 'utf8');
}

function cleanup() {
    setTimeout(() => {
        logger.error(`Таймаут commit-msg hook ${timeToWrite}${FORMAT.NEW_LINE}`);

        process.exit(1);
    }, timeToWrite);
}

/**
 * Проверка сообщения коммита
 */
function checkCommitMessage() {
    let hasError = false;
    let commitMessage = fs.readFileSync(commitEditMsgFile, 'utf8');

    const commitMessageLines = commitMessage.split(EOL);

    const firstLineRegExp = /^\s?#[A-Z]+-\d+\s\[(feature|hotfix|bugfix|refactor|doc)\]\s.+(\n\n(\s{4}(\+|-|!|\*)\s.+\n?)*)*$/g;

    commitMessage = commitMessageLines.join('');
    if (!firstLineRegExp.test(commitMessageLines.join(''))) {
        logger.error('');
        console.error(`⚠️  \x1b[1;31mТекст коммита не соответствует шаблону.${FORMAT.RESET}\n`);
        console.error(`${COLORS.BGMAGENTA}${commitMessage}${FORMAT.RESET}\n`);
        console.log('\x1b[32mПримеры описаний коммитов:');
        console.log(`\x1b[32m--------------------------------------------------------|${FORMAT.RESET}`);
        console.log('#CBDL-635 [feature] добавляет возможность авторизации  \x1b[32m |');
        console.log(`\x1b[32m--------------------------------------------------------|${FORMAT.RESET}`);
        console.log('#CBDL-677 [bugfix] фиксит баг с нерабочей сортировкой  \x1b[32m |');
        console.log(`\x1b[32m--------------------------------------------------------|${FORMAT.RESET}`);
        console.log('[refactor] выносит часть кода в компонент Aside        \x1b[32m |');
        console.log(`\x1b[32m--------------------------------------------------------|${FORMAT.RESET}`);
        console.log(`#CBDL-700 [feature] Вносит правки в страницу логина    \x1b[32m |${FORMAT.RESET}
    + лейбл с подсказкой                               \x1b[32m |${FORMAT.RESET}
    - браузерную валидацию                             \x1b[32m |${FORMAT.RESET}
    ! неработающий сабмит формы                        \x1b[32m |${FORMAT.RESET}
    * порядок полей в форме                            \x1b[32m |`);
        console.log(`\x1b[32m--------------------------------------------------------|${FORMAT.RESET}\n`);
        console.log(`Подробнее: ${COLORS.BLUE}${FORMAT.UNDERLINE}https://kb.akbars.ru/pages/viewpage.action?pageId=73574927${FORMAT.RESET}`);
        logger.default('');
        hasError = true;
    }

    return hasError;
}

async function main() {
    let task = '', taskType = '', branchName = await getCurrentBranch();
    let hasCommitMessageError = false;
    let hasBranchError = false;
    hasCommitMessageError = checkCommitMessage();

    try {
        taskType = await getTaskTypeFromBranch(branchName);

        if (taskType) {
            writeTaskTypeToTitleInFile(taskType, commitEditMsgFile);
        }
    } catch ({ message }) {
        console.log(message);
        hasBranchError = true;
    }

    try {
        task = await getTaskFromBranch(branchName);

        if (task) {
            writeTaskToTitleInFile(task, commitEditMsgFile);
        }
    } catch ({ message }) {
        console.log(message);
        hasBranchError = true;
    }

    if (hasCommitMessageError || hasBranchError) {
        process.exit(1);
    }

    process.exit(0);
}

cleanup();
main();
